/* Call Windows NT 3.x linker.
   Copyright (C) 1994 Free Software Foundation, Inc.
   Contributed by Douglas B. Rupp (drupp@cs.washington.edu).

This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include "config.h"
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>
#include <process.h>
#include <stdarg.h>
#if 0
#include <varargs.h>
#endif
#include <fcntl.h>

#ifdef __GNUC__
#define spawnvp _spawnvp
#define stat _stat
#define strdup _strdup
#endif __GNUC__

char *standard_framework_dirs[] = {
	"/LocalLibraray/Frameworks/", 
	"/NextLibrary/Frameworks/",
	NULL
};

char **framework_dirs = NULL;
unsigned long nframework_dirs = 0;

char **framework_names = NULL;
unsigned long nframework_names = 0;

/* These can be set by command line arguments */
char *linker_path = 0;
int verbose = 0;

int link_arg_max = -1;
char **link_args = (char **) 0;
int link_arg_index = -1;

char *search_dirs = 0;

static int is_regular_file (char *name);

static void
addarg (str)
     char *str;
{
  int i;

  if (++link_arg_index >= link_arg_max)
    {
      char **new_link_args
	= (char **) calloc (link_arg_max + 1000, sizeof (char *));

      for (i = 0; i <= link_arg_max; i++)
	new_link_args [i] = link_args [i];

      if (link_args)
	free (link_args);

      link_arg_max += 1000;
      link_args = new_link_args;
    }

  link_args [link_arg_index] = str;
}

static char *
locate_file (file_name, path_val)
     char *file_name;
     char *path_val;
{
  char buf [1000];
  int file_len = strlen (file_name);
  char *end_path = path_val + strlen (path_val);
  char *ptr;

  /* Handle absolute pathnames */
  if (file_name [0] == '/' || file_name [0] == DIR_SEPARATOR
      || isalpha (file_name [0]) && file_name [1] == ':')
    {
      strncpy (buf, file_name, sizeof buf);
      buf[sizeof buf - 1] = '\0';
      if (is_regular_file (buf))
	return strdup (buf);
      else
	return 0;
  }

  if (! path_val)
    return 0;

  for (;;)
    {
      for (; *path_val == PATH_SEPARATOR ; path_val++)
	;
      if (! *path_val)
	return 0;

      for (ptr = buf; *path_val && *path_val != PATH_SEPARATOR; )
	*ptr++ = *path_val++;

      ptr--;
      if (*ptr != '/' && *ptr != DIR_SEPARATOR)
	*++ptr = DIR_SEPARATOR;

      strcpy (++ptr, file_name);

      if (is_regular_file (buf))
	return strdup (buf);
    }

  return 0;
}

static char *
expand_lib (name)
     char *name;
{
  char *lib, *lib_path;

  lib = malloc (strlen (name) + 6);
  strcpy (lib, "lib");
  strcat (lib, name);
  strcat (lib, ".a");


  lib_path = locate_file (lib, search_dirs);
  if (!lib_path)
    {
#if defined (_WIN32) && defined (NEXT_PDO)
      char *lib2;
      lib2 = (char *) malloc (strlen (name) + 5);
      strcpy (lib2, name);
      strcat (lib2, ".lib");
      lib_path = locate_file (lib2, search_dirs);
      if (!lib_path)
	{
	  fprintf (stderr, "Couldn't locate library: %s or %s\n", lib, lib2);
	  free (lib2);
	  exit (-1);
	}
      free (lib2);
#else
      fprintf (stderr, "Couldn't locate library: %s\n", lib);
      exit (-1);
#endif /* _WIN32 */
    }

  return lib_path;



}

static int
is_regular_file (name)
     char *name;
{
  int ret;
  struct _stat statbuf;

  ret = stat(name, &statbuf);
  return !ret && S_ISREG (statbuf.st_mode);
}

static int
is_directory (name)
     char *name;
{
  int ret;
  struct _stat statbuf;

  ret = stat(name, &statbuf);
  return !ret && S_ISDIR (statbuf.st_mode);
}

static void
process_args (p_argc, argv)
     int *p_argc;
     char *argv[];
{
  int i, j;

  for (i = 1; i < *p_argc; i++)
    {
      /* -v turns on verbose option here and is passed on to gcc */
      if (! strcmp (argv [i], "-v"))
	verbose = 1;
    }
}




/*
 * search_for_framework() takes name and tries to open a file with that name
 * in the -F search directories and in the standard framework directories.  If
 * it is sucessful it returns a pointer to the file name indirectly through
 * file_name and the open file descriptor indirectly through fd.
 */
static
void
search_for_framework(
char *name,
char **file_name,
int *fd)
{
    unsigned long i;

	*fd = -1;
	for(i = 0; i < nframework_dirs ; i++){
		*file_name = (char*)malloc (strlen(framework_dirs[i]) + strlen(name) + 6);
		// First try with a .lib on it
		sprintf (*file_name, "%s%s%s.lib", framework_dirs[i], "/", name);
	    if((*fd = open(*file_name, O_RDONLY, 0)) != -1)
		{
			close(*fd);
			break;
		}
		// Then try with nothing.
		sprintf (*file_name, "%s%s%s.", framework_dirs[i], "/", name);
	    if((*fd = open(*file_name, O_RDONLY, 0)) != -1)
		{
			close(*fd);
			break;
		}
	    free(*file_name);
	}
	if(*fd == -1){
	    for(i = 0; standard_framework_dirs[i] != NULL ; i++){
		*file_name = (char*)malloc (strlen(standard_framework_dirs[i]) + strlen(name) + 5);
		// First try with a .lib on it
		sprintf (*file_name, "%s%s.lib", standard_framework_dirs[i], name);
		if((*fd = open(*file_name, O_RDONLY, 0)) != -1)
		{
		    close(*fd);
			break;
		}
		// Then try with nothing.
		sprintf (*file_name, "%s%s.", standard_framework_dirs[i], name);
		if((*fd = open(*file_name, O_RDONLY, 0)) != -1)
		{
		    close(*fd);
			break;
		}
		free(*file_name);
	    }
	}
}



main (argc, argv)
     int argc;
     char *argv[];
{
  int i;
  int done_an_ali = 0;
  int file_name_index;
  int using_codeview = 0;	// if 1 then -debugtype:both otherwise -debugtype:coff
  int undefined_flag = 0; /* records the value of the -undefined switch */
  char *pathval = getenv ("PATH");
  char *spawn_args [5];
  char *tmppathval = malloc (strlen (pathval) + 3);

#if defined (_WIN32) && defined (NEXT_PDO)
  char *next_root = getenv( "NEXT_ROOT" );

  if (next_root && *next_root)
    {
      if (next_root[strlen(next_root)-1] == '/')
	next_root[strlen(next_root)-1] = '\0';
      for (i = 0; standard_framework_dirs[i]; i++)
	{
	  char* new_fname =
	    (char *) malloc (strlen (next_root)
			     + strlen (standard_framework_dirs[i]) + 1);
	  sprintf (new_fname, "%s%s", next_root, standard_framework_dirs[i]);
	  if (new_fname && *new_fname)
	    standard_framework_dirs[i] = new_fname;
	}
    }
#endif

  /* Set up the search dirs variable */
  search_dirs = (char*)malloc( strlen( getenv ("LIB") ) + 1 );
  strcpy( search_dirs, getenv ("LIB") );
  strcat( search_dirs, "\0" );

  strcpy (tmppathval, ".;");
  pathval = strcat (tmppathval, pathval);

  process_args (&argc , argv);

  linker_path = locate_file ("link32.exe", pathval);
  if (!linker_path)
    {
      linker_path = locate_file ("link.exe", pathval);
      if (!linker_path)
	{
	  fprintf (stderr, "Couldn't locate link32 or link\n");
	  exit (-1);
	}
    }

  addarg (linker_path);

  for (i = 1; i < argc; i++)
    {
      int arg_len = strlen (argv [i]);

      if (!strcmp (argv [i], "-o"))
	{
	  char *buff, *ptr;
	  int out_len;

	  i++;
	  out_len = strlen (argv[i]) + 10;
	  buff = malloc (out_len);
	  strcpy (buff, "-out:");
	  strcat (buff, argv[i]);
	  ptr = strstr (buff, ".exe");
	  if( ptr == NULL )
	  	ptr = strstr( buff, ".dll" );
	  if (ptr == NULL || strlen (ptr) != 4)
	    strcat (buff, ".exe");
	  addarg (buff);
	}
    else if (!strcmp (argv [i], "-g"))
        {
	  /* if we get a -g flag, then that means that we should
	     generate CodeView formated debugging information */
	  using_codeview = 1;
        }
      else if (arg_len > 2 && !strncmp (argv [i], "-L", 2))
	{
	  char *nbuff, *sdbuff;
	  int j, new_len, search_dirs_len;

	  new_len = strlen (&argv[i][2]);
	  search_dirs_len = strlen (search_dirs);

	  nbuff = malloc (new_len + 1);
	  strcpy (nbuff, &argv[i][2]);

	  for (j = 0; j < new_len; j++)
	    if (nbuff[j] == '/') nbuff[j] = DIR_SEPARATOR;

	  sdbuff = malloc (search_dirs_len + new_len + 2);
	  strcpy (sdbuff, search_dirs);
	  sdbuff[search_dirs_len] = PATH_SEPARATOR;
	  sdbuff[search_dirs_len+1] = 0;
	  strcat (sdbuff, nbuff);

	  if (search_dirs)
	    free (search_dirs);

	  search_dirs = sdbuff;
	}

      else if (arg_len > 2 && !strncmp (argv [i], "-l", 2))
	addarg (expand_lib (&argv[i][2]));
#if defined (_WIN32) && defined (NEXT_PDO)
	else if (!strcmp (argv [i], "-framework"))
	{
		if (++i >= argc)
		{
			fprintf (stderr, "Error: -framework: argument missing\n");
    		exit (-1);
		}
	
		framework_names = (char**)realloc (framework_names, (nframework_names + 1) * sizeof (char*));
		framework_names[nframework_names++] = argv[i];
	}
      else if (!strncmp (argv [i], "-F", 2))
	{
		if (argv[i][2] == '\0')
		{
			fprintf (stderr, "Error: -F: directory name missing\n");
    		exit (-1);
		}
	
		framework_dirs = (char**)realloc (framework_dirs, (nframework_dirs + 1) * sizeof (char*));
		framework_dirs[nframework_dirs++] = &(argv[i][2]);
		if (!is_directory (&(argv[i][2])))
		{
			fprintf (stderr, "Warning: -F: directory name (%s) does not exist\n", &(argv[i][2]));
    		// exit (-1);
		}
	}
      else if (!strcmp (argv[i], "-undefined"))
	{
	  if (++i >= argc)
	    {
	      fprintf (stderr, "Error: Argument to -undefined missing\n");
	      exit (-1);
	    }
	  if (!strcmp (argv[i], "warning") || !strcmp (argv[i], "suppress"))
	    {
	      if (!undefined_flag)
		{
		  undefined_flag = 1;
		  addarg ("-force:unresolved");
		}
	      else if (undefined_flag != 1)
		{
		  fprintf (stderr,
			   "Error: Multiple conflicting -undefined flags\n");
		  exit (-1);
		}
	    }
	  else if (!strcmp (argv[i], "error"))
	    {
	      if (undefined_flag > 0)
		{
		  fprintf (stderr,
			   "Error: Multiple conflicting -undefined flags\n");
		  exit (-1);
		}
	      undefined_flag = -1;
	    }
	  else
	    {
	      fprintf (stderr,
		       "Error: unrecognized option \"-undefined %s\"\n",
		       argv[i]);
	      exit (-1);
	    }
	}
#endif
    else if (!strcmp (argv [i], "-v")) 
      ;
    else
      addarg (argv [i]);
    }
/* This is important.  If using_codeview is 1 then we should tell
   the linker to generate both CodeView and COFF formats.  If using_codeview
   isn't 1, then we still need to generate COFF format debugging 
   info so that when collect2 runs nm on the executable it will
   find symbols and can properly do the second link! */
  if( using_codeview )
    addarg( "-debug -debugtype:both" );
  else
    addarg( "-debug -debugtype:coff" );

/* Now try to find all the frameworks that were specified using -framework. */
	{
		char *type;
		char *p;
		char *file_name;
		int count;
		int fd = -1;			// not used, only for return value;


		for (count = 0; count < nframework_names; count++)
		{
			fd = -1;
			type = strrchr(framework_names[count], ',');
			if(type != NULL && type[1] != '\0'){
			*type = '\0';
			type++;
			p = (char*)malloc ((strlen(framework_names[count])*2) + 11 + strlen(type) + 2);
			sprintf (p, "%s%s%s%s", framework_names[count], ".framework/", framework_names[count], type);
			search_for_framework(p, &file_name, &fd);
			if(fd == -1)
			{
			    fprintf(stderr, "can't locate framework link library for: -framework %s,%s "
				    "using suffix %s\n", framework_names[count], type, type);
			}
			else
				addarg (file_name);
			}
			else
			type = NULL;
			if(fd == -1){
			p = (char*)malloc ((strlen(framework_names[count])*2) + 11 + 2);
			sprintf (p, "%s%s%s", framework_names[count], ".framework/", framework_names[count]);
			search_for_framework(p, &file_name, &fd);
			}
			if(fd == -1)
			{
				if(type != NULL)
				{
				    fprintf(stderr, "can't locate framework link library for: -framework %s,%s\n",
					  framework_names[count], type);
					exit(-1);
				}
				else
				{
				    fprintf(stderr, "can't locate framework link library for: -framework %s\n", framework_names[count]);
					exit(-1);
				}
			}
			else
				addarg (file_name);
		}
	}
	



  addarg (NULL);

  if (verbose)
    {
      int i;

      for (i = 0; i < link_arg_index; i++)
	printf ("%s ", link_args [i]);
      putchar ('\n');
    }

  if (spawnvp (P_WAIT, linker_path, (const char * const *)link_args) != 0)
    {
//      fprintf (stderr, "Error executing %s\n", link_args[0]);
      exit (-1);
    }

  exit (0);
}

